##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os

from zope.interface import implements

from interfaces import IXSLFOConverter
from config import supported_formats
from html2fo import HTML2FOConverter
from fo2xfc import (xfc_available, RTFConverter, ODTConverter, 
                    WMLConverter, DOCXConverter)
from fo2pdf import xinc_available, PDFConverter
from fo2fop import fop_available, FOPPDFConverter


class Converter(object):
    """High-level OO interface to XSL-FO conversions """

    implements(IXSLFOConverter)

    def __init__(self, filename, encoding='utf-8', cleanup=False, verbose=False):
        self.filename = filename
        self.encoding = encoding
        self.fo_filename = None        
        self.cleanup = cleanup
        self.verbose = verbose

    def convert(self, format, output_filename=None):

        if format == 'fo':
            if self.fo_filename is None:
                self.fo_filename = HTML2FOConverter().convert(self.filename, self.encoding)
            return self.fo_filename

        if not format in supported_formats:
            raise ValueError('Unsupported format: %s' % format)

        if self.fo_filename is None:
            self.fo_filename = HTML2FOConverter().convert(self.filename, self.encoding)

        if format == 'pdf':
            if not xinc_available:
                raise RuntimeError("The external XINC converter isn't available")
            output_filename = PDFConverter().convert(self.fo_filename, output_filename=output_filename)

        elif format == 'pdf2':
            if not fop_available:
                raise RuntimeError("The external FOP converter isn't available")
            output_filename = FOPPDFConverter().convert(self.fo_filename, output_filename=output_filename)

        else:
            if not xfc_available:
                raise RuntimeError("The external XFC converter isn't available")

            map = {'odt' : ODTConverter,
                   'rtf' : RTFConverter,
                   'docx' : DOCXConverter,
                   'wml' : WMLConverter }
            output_filename = map[format]().convert(self.fo_filename, output_filename=output_filename)

        return output_filename

    def __call__(self, *args, **kw):
        return self.convert(*args, **kw)

    def __del__(self):
        """ House-keeping """

        if self.cleanup:
            if self.fo_filename:
                os.unlink(self.fo_filename)




##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

"""
Tests, tests, tests.........
"""

import os
import sys
import unittest

from zope.interface.verify import verifyClass
from zopyx.convert.interfaces import IFOConverter, IHTML2FOConverter, IXSLFOConverter

from zopyx.convert import HTML2FOConverter
from zopyx.convert import RTFConverter, WMLConverter, DOCXConverter, ODTConverter
from zopyx.convert import PDFConverter, Converter


html1 = os.path.join(os.path.dirname(__file__), 'data', 'test1.html')
html2 = os.path.join(os.path.dirname(__file__), 'data', 'test2.html')

class ConverterTests(unittest.TestCase):

    def testInterfaces(self):
        verifyClass(IHTML2FOConverter, HTML2FOConverter)
        verifyClass(IFOConverter, RTFConverter)
        verifyClass(IFOConverter, ODTConverter)
        verifyClass(IFOConverter, DOCXConverter)
        verifyClass(IFOConverter, WMLConverter)
        verifyClass(IXSLFOConverter, Converter)

class RegressionTests(unittest.TestCase):

    def _test(self, converter):

        for filename in (html1, html2):
            fo_filename = HTML2FOConverter().convert(filename)
            try:
                output_filename = converter().convert(fo_filename)
            except: 
                print >>sys.stderr, 'Error converting %s' % filename
                raise

    def testPDF(self):
        return self._test(PDFConverter)

    def testRTF(self):
        return self._test(RTFConverter)

    def testODT(self):
        return self._test(ODTConverter)

    def testWML(self):
        return self._test(WMLConverter)

    def testDOCX(self):
        return self._test(DOCXConverter)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(ConverterTests))
    suite.addTest(makeSuite(RegressionTests))
    return suite




##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

from zope.interface import Interface

class IFOConverter(Interface):

    def convert(fo_filename):
        """ Convert a XSL-FO file (stored on the filesystem) to some output format.
            Returns the filename of the output file. 
        """


class IHTML2FOConverter(Interface):
        
    def convert(html_filename):
        """ Convert a HTML file (stored on the filesystem to XSL-FO.    
        """


class IXSLFOConverter(Interface):
    """ A converter that converts HTML to some output format using
        XSL-FO.
    """

    def __init__(filename, encoding='utf-8', cleanup=False, verbose=False):
        """ Initialize a converter with the 'filename' of a HTML file """

    def convert(format, output_filename=None):
        """ Convert HTML to output 'format' """



##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import sys
import tempfile

from subprocess import Popen, PIPE
from logger import LOG

win32 = (sys.platform=='win32')


def newTempfile(suffix=''):
    return tempfile.mktemp(suffix=suffix)


def runcmd(cmd):                
    """ Execute a command using the subprocess module """

    if win32:

        cmd = cmd.replace('\\', '/')
        s = Popen(cmd, shell=False)
        s.wait()
        return 0, ''

    else:      

        stdin = open('/dev/null')
        stdout = stderr = PIPE

        p = Popen(cmd, 
                  shell=True,
                  stdin=stdin,
                  stdout=stdout,
                  stderr=stderr,
                  )

        status = p.wait()
        stdout_ = p.stdout.read().strip()
        stderr_ = p.stderr.read().strip()

        if stdout_:
            LOG.info(stdout_)
        if stderr_:
            LOG.info(stderr_)
        return status, stdout_ + stderr_




##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import tempfile
from optparse import OptionParser
from convert import Converter
from config import supported_formats

from zopyx.convert import availableFormats

def convert(options, args):

    if options.test_mode:

        import pkg_resources
        print 'Entering testmode'

        for fn in ('test1.html', 'test2.html', 'test3.html'):
            tmpf = tempfile.mktemp() 
            open(tmpf + '.html', 'wb').write(pkg_resources.resource_string('zopyx.convert.tests.data', fn))

            for format in availableFormats():
                print '%s: %s.html -> %s.%s' % (format, tmpf, tmpf, format)
                C = Converter(tmpf + '.html', verbose=True)
                output_filename = C(format, output_filename=tmpf + '.' + format)

    elif options.show_converters:
        print 'Available converters: %s' % ','.join(availableFormats())

    else:

        for fn in args:
            C = Converter(fn, verbose=options.verbose)
            output_filename = C(options.format, output_filename=options.output_filename)
            print 'Generated file: %s' % output_filename
   

def main():

    parser = OptionParser()
    parser.add_option('-v', '--verbose', dest='verbose', action='store_true',
                      default=False, help='verbose on')

    parser.add_option('-f', '--format', dest='format',
                      help='|'.join(supported_formats))

    parser.add_option('-l', '--list-converters', action='store_true', dest='show_converters',
                      default=False, help='show all available converters')

#    parser.add_option('-c', '--check-converters', action='store_true', dest='check_converters',
#                      default=False, help='check all available converters')

    parser.add_option('-o', '--output', dest='output_filename',
                      help='output filename')

    parser.add_option('-t', '--test', dest='test_mode', action='store_true',
                      help='test converters')
    (options, args) = parser.parse_args()
    convert(options, args)


if __name__ == '__main__':
    main()



##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

from interfaces import IFOConverter, IHTML2FOConverter, IXSLFOConverter

from convert import Converter
from html2fo import HTML2FOConverter
from fo2pdf import PDFConverter, xinc_available
from fo2fop import FOPPDFConverter, fop_available
from fo2xfc import DOCXConverter, RTFConverter, ODTConverter, WMLConverter, xfc_available


def availableFormats():

    formats = []
    if xinc_available:
        formats.append('pdf')
    if fop_available:
        formats.append('pdf2')
    if xfc_available:
        formats.extend(['rtf', 'docx', 'odt', 'wml'])

    return formats



##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
import sys

from util import newTempfile, runcmd
from logger import LOG

from zope.interface import implements
from interfaces import IFOConverter

fop_home = os.environ.get('FOP_HOME')

def _check_fop():
    if not fop_home:
        LOG.error('$FOP_HOME not set')
        return False

    if not os.path.exists(fop_home):
        LOG.error('$FOP_HOME=%s does not exist' % fop_home)
        return False

    return True        

fop_available = _check_fop()

class FOPPDFConverter(object):

    implements(IFOConverter)

    def convert(self, fo_filename, output_filename=None):
        """ Convert a FO file to PDF using FOP"""

        if not output_filename:
            output_filename = newTempfile(suffix='.pdf')

        if sys.platform == 'win32':
            cmd = '%s\\fop.bat -fo "%s" -pdf "%s"' % (fop_home, fo_filename, output_filename)
        else:
            cmd = '"%s/fop" -fo "%s" -pdf "%s"' % (fop_home, fo_filename, output_filename)

        status, output = runcmd(cmd)
        if status != 0:
            raise RuntimeError('Error executing: %s' % cmd)

        return output_filename




##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
from setuptools import setup, find_packages


CLASSIFIERS = [
    'Programming Language :: Python',
]

version_file = os.path.join('src', 'zopyx', 'convert', 'version.txt')
version = open(version_file).read().strip()

readme_file = os.path.join('src', 'zopyx', 'convert', 'README.txt')
desc = open(readme_file).read().strip()

changes_file = os.path.join('src', 'zopyx', 'convert', 'CHANGES.txt')
changes = open(changes_file).read().strip()

long_description = desc + '\n\nChanges:\n========\n\n' + changes


setup(name='zopyx.convert',
      version=version,
      license='ZPL (see LICENSE.txt)',
      author='Andreas Jung',
      author_email='info@zopyx.com',
      maintainer='Andreas Jung',
      maintainer_email='info@zopyx.com',
      classifiers=CLASSIFIERS,
      url='http://cheeseshop.python.org/pypi/zopyx.convert',
      description='A Python interface to XSL-FO libraries (Conversion HTML to PDF, RTF, DOCX, WML and ODT)',
      long_description=long_description,
      packages=['zopyx', 'zopyx.convert'],
      package_dir = {'': 'src'},
      include_package_data = True,
      zip_safe=False,
      install_requires=['setuptools'],
      namespace_packages=['zopyx'],
      entry_points={'console_scripts': ['xslfo-convert = zopyx.convert.cli:main',]},
      )                       



##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
import re

from config import java
from util import newTempfile, runcmd
from elementtree.ElementTree import parse, tostring, SubElement        
from htmlentitydefs import name2codepoint
from BeautifulSoup import BeautifulSoup

from zope.interface import implements
from interfaces import IHTML2FOConverter

dirname = os.path.dirname(__file__)


def tidyhtml(filename, encoding):

    html = open(filename, 'rb').read()

    # use BeautifulSoup for performing HTML checks
    # and conversion to XHTML
    soup = BeautifulSoup(html)

    # check if all image files exist
    for img in soup.findAll('img'):
        src = img['src']
        if not os.path.exists(src):
            raise IOError('No image file found: %s' % src)

    html = str(soup.prettify())

    # replace the HTML tag
    html = '<html xmlns="http://www.w3.org/1999/xhtml">' + \
            html[html.find('<html') + 6:]
    # add the XML preamble
    html = '<?xml version="1.0" ?>\n' + html

    # replace all HTML entities with numeric entities
    def handler(mo):
        """ Callback to convert entities """
        e = mo.group(1)
        v = e[1:-1]
        if not v.startswith('#'):
            codepoint =  name2codepoint.get(v)
            return codepoint and '&#%d;' % codepoint or ''
        else:
            return e
    
    entity_reg = re.compile('(&.*?;)')
    html = entity_reg.sub(handler, html)

    filename = newTempfile()
    open(filename, 'wb').write(str(html))
    return filename


class HTML2FOConverter(object):  

    implements(IHTML2FOConverter)

    def convert(self, filename, encoding='utf-8', language='en', country='en', tidy=True):
        """ Convert a HTML file stored as 'filename' to
            FO using CSS2XSLFO.
        """

        if tidy:
            filename = tidyhtml(filename, encoding)

        fo_filename = newTempfile(suffix='.fo')
        csstoxslfo = os.path.abspath(os.path.join(dirname, 'lib', 'csstoxslfo', 'css2xslfo.jar'))
        if not os.path.exists(csstoxslfo):
            raise IOError('%s does not exist' % csstoxslfo)

        cmd = '"%s"' % java + \
              ' -Duser.language=en -Xms128m -Xmx128m -jar "%(csstoxslfo)s" "%(filename)s" -fo "%(fo_filename)s" country=%(country)s language=%(language)s' % vars()
        status, output = runcmd(cmd)
        if status != 0:
            raise RuntimeError('Error executing: %s' % cmd)

        # remove tidy-ed file
        os.unlink(filename)

        # remove some stuff from the generated FO file causing
        # some conversion trouble either with XINC or XFC

        E = parse(fo_filename)

        ids_seen = list()
        for node in E.getiterator():
            get = node.attrib.get

            # ensure that ID attributes are unique
            node_id = get('id')
            if node_id is not None:
                if node_id in ids_seen:
                    del node.attrib['id']
                ids_seen.append(node_id)

            for k, v in (('footnote', 'reset'), 
                         ('unicode-bidi', 'embed'), 
                         ('writing-mode', 'lr-tb'), 
                         ('font-selection-strategy', 'character-by-character'), 
                         ('line-height-shift-adjustment', 'disregard-shifts'), 
                         ('page-break-after', 'avoid'), 
                         ('page-break-before', 'avoid'), 
                         ('page-break-inside', 'avoid')):

                value = get(k)
                if value == v:
                    del node.attrib[k]

            for attr in ('margin-left', 'margin-right', 'margin-top', 'margin-bottom',
                         'padding-left', 'padding-right', 'padding-top', 'padding-bottom'):

                value = get(attr)
                if value == '0':
                    node.attrib[attr] = '0em'

            if get('page-break-after') == 'always':
                del node.attrib['page-break-after']
                node.attrib['break-after'] = 'page'

            if get('text-transform'):
                del node.attrib['text-transform']

            value = get('white-space')
            if value == 'pre':
                del node.attrib['white-space']
                node.text = '\n' + node.text.lstrip()
                for k,v in  {'white-space-treatment' : 'preserve',
                             'white-space-collapse' : 'false',
                             'wrap-option' : 'no-wrap',
                             'linefeed-treatment' : 'preserve' }.items():
                    node.attrib[k] = v
                
        fo_text = tostring(E.getroot())
        fo_text = fo_text.replace('<ns0:block ' , '<ns0:block margin-top="0" margin-bottom="0" ')  # avoid a linebreak through <li><p> (XFC)
#        fo_text = fo_text.replace('<ns0:block/>', '') # causes a crash with XINC    
        fo_text = fo_text.replace('<ns0:block margin-top="0" margin-bottom="0" />', '') 


        open(fo_filename, 'wb').write(fo_text)
        return fo_filename




##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
import sys

from util import newTempfile, runcmd
from logger import LOG

from zope.interface import implements
from interfaces import IFOConverter

xinc_home = os.environ.get('XINC_HOME')

def _check_xinc():
    if not xinc_home:
        LOG.error('$XINC_HOME not set')
        return False

    if not os.path.exists(xinc_home):
        LOG.error('$XINC_HOME=%s does not exist' % xinc_home)
        return False

    return True        

xinc_available = _check_xinc()

class PDFConverter(object):

    implements(IFOConverter)

    def convert(self, fo_filename, output_filename=None):
        """ Convert a FO file to PDF using XINC """

        if not output_filename:
            output_filename = newTempfile(suffix='.pdf')

        if sys.platform == 'win32':
            cmd = '%s\\xinc.exe -fo "%s" -pdf "%s"' % (xinc_home, fo_filename, output_filename)
        else:
            cmd = '"%s/bin/unix/xinc" -fo "%s" -pdf "%s"' % (xinc_home, fo_filename, output_filename)

        status, output = runcmd(cmd)
        if status != 0:
            raise RuntimeError('Error executing: %s' % cmd)

        return output_filename




##########################################################################
# zopyx.convert - XSL-FO related functionalities
#
# (C) 2007, ZOPYX Ltd & Co. KG, Tuebingen, Germany
##########################################################################

import os
import sys

from util import newTempfile, runcmd
from logger import LOG

from zope.interface import implements
from interfaces import IFOConverter

xfc_dir = os.environ.get('XFC_DIR')

def _check_xfc():

    if not xfc_dir:
        LOG.error('$XFC_DIR not set')
        return False

    if not os.path.exists(xfc_dir):
        LOG.error('$XFC=%s does not exist' % xfc_dir)
        return False

    return True

xfc_available = _check_xfc()


def fo2xfc(fo_filename, format='rtf', output_filename=None):
    """ Convert a FO file to some format support 
        through XFC-4.0.
    """

    if not format in ('rtf', 'docx', 'wml', 'odt'):
        raise ValueError('Unsupported format: %s' % format)

    if not output_filename:
        output_filename = newTempfile(suffix='.%s' % format)

    if sys.platform == 'win32':
        cmd = '"%s\\fo2%s.bat"  "%s" "%s"' % (xfc_dir, format, fo_filename, output_filename) 
    else:	
        cmd = '"%s/fo2%s" "%s" "%s"' % (xfc_dir, format, fo_filename, output_filename)

    status, output = runcmd(cmd)
    if status != 0:
        raise RuntimeError('Error executing: %s' % cmd)

    return output_filename


class RTFConverter(object):

    implements(IFOConverter)
    def convert(self, fo_filename, output_filename=None):
        return fo2xfc(fo_filename, 'rtf', output_filename)

class ODTConverter(object):

    implements(IFOConverter)
    def convert(self, fo_filename, output_filename=None):
        return fo2xfc(fo_filename, 'odt', output_filename)

class WMLConverter(object):

    implements(IFOConverter)
    def convert(self, fo_filename, output_filename=None):
        return fo2xfc(fo_filename, 'wml', output_filename)

class DOCXConverter(object):

    implements(IFOConverter)
    def convert(self, fo_filename, output_filename=None):
        return fo2xfc(fo_filename, 'docx', output_filename)

